//---------------------------------------------------------------------------
// File: TMainForm.cpp
//
//  TDx_Draw_Library
//  ExampleQ - Converted DirectX Code - Donuts
//
// Date: 21st June, 2002
// Author: Darren John Dwyer
// Copyright: (c) 2002 Darren John Dwyer, All Rights Reserved.
//
// Description:
//
// This example is a fairly straightforward conversion from
// DirectDraw's Example4 Demo.
//
// It uses the file "all.bmp" supplied with the DirectX SDK.
//
// This application has been "translated" from the Original DirectX Code.
// The code is very similar to the "Example4" code supplied by MS, just
// converted so it works with the TDx_Draw_Library and C++ Builder.
//
// It consists of the following code sections :-
//
// FormCreate() - sets up DirectDraw, loads the images, sets palettes, etc.
// FormDestroy() - cleans up.
// FormKeyDown() - handles keystrokes.
// FormMouseMove() - updates the mouse cursor display
// FormResize() - updates the size of the status list box (when errors occur)
//
// OnMinimize() - stops animation
// OnRestore() - restores the cooperative level and continues animation
// OnIdle() - calls a method called UpdateFrame() to paint each frame.
//
// UpdateFrame() - paints each frame.
//
// PrimarySurface::SurfaceLost() - handles restoration of surface data.
// Dx_Draw1::OnError() - handles all DirectDraw errors.
// Dx_Draw1::OnEnumDisplayModes() - determines available modes.
//
//---------------------------------------------------------------------------
// SPECIAL NOTE:
//
// This application uses 640x480x8 normally.
//
// In order to save a snapshot of the screen, I've modified it slightly
// so that the snapshot palette is saved too (using 640x480x32).
//
// When doing this, note that you don't actually need to interface with
// DirectDraw palette functions. A nice bonus.
//
// To switch between the two, recompile the code after changing 'BITS'
//
// When 'BITS' is 8, no palette information appears to be saved by BCB.
// When 'BITS' is 16, no palette information appears to be saved by BCB.
// When 'BITS' is 24, no palette information appears to be saved by BCB.
// When 'BITS' is 32, the palette information is saved correctly by BCB.
//
//
//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop

#include "TMainForm.h"
#include "mmsystem.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
//---------------------------------------------------------------------------
#pragma link "TDx_Draw"
#pragma link "TDx_DrawSurface"
#pragma link "TDDColorKey"
#pragma link "TDDSCaps"
#pragma link "TDDSurfaceDesc"
#pragma link "TDx_DrawPalette"
#pragma link "TDDCaps"
#pragma link "tddcolorkey"
#pragma link "tddscaps"
#pragma link "tddsurfacedesc"
#pragma link "tdx_draw"
#pragma link "tdx_drawpalette"
#pragma link "tdx_drawsurface"
#pragma link "TDx_Draw_Library_Install"
#pragma resource "*.dfm"
//---------------------------------------------------------------------------
TMainForm *MainForm;

//---------------------------------------------------------------------------
// 'BITS' can normally be:  8, 16, 24 and 32
//---------------------------------------------------------------------------
dword BITS = 8;

//---------------------------------------------------------------------------
__fastcall TMainForm::TMainForm(TComponent* Owner)
    : TForm(Owner)
{
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::FormCreate(TObject *Sender)
{
//
// FormCreate:
//
// Sets up the current form for a Full-Screen / Exclusive Mode application.
// Sets up the required DirectDraw objects.
// Sets up application-specific events

// setup Form
BorderStyle = bsNone;
Screen->Cursor = crNone;
WindowState = wsMaximized;

// setup Dx_Draw1
bool ok = true;
ok &= Dx_Draw1->Create( NULL );

// check if there is a 640x480x8 mode on the system.
// note: ModeExists and
//       NumDisplayModes
//       are updated inside Dx_Draw1::OnEnumDisplayModesCallback
ModeExists = false;
NumDisplayModes = 0;
Dx_Draw1->EnumDisplayModes( 0, NULL, NULL );

// if the mode was detected in the OnEnumDisplayModesCallback event
if (ModeExists)
    {
    // setup DirectX
    ok &= Dx_Draw1->SetCooperativeLevel( Application->Handle, DDSCL_EXCLUSIVE | DDSCL_FULLSCREEN );
    ok &= Dx_Draw1->SetDisplayMode( 640, 480, BITS, 0, 0 );

    // nb: for some reason or other, DirectDraw seems to forget that you are still in exclusive mode,
    //     so, little choice but to re-apply exclusive mode...
    ok &= Dx_Draw1->SetCooperativeLevel( Application->Handle, DDSCL_EXCLUSIVE | DDSCL_FULLSCREEN );

    // setup the Primary Surface
    ok &= PrimarySurface->Create( PrimarySurfaceDesc, Dx_Draw1 );
    // put the Surface Description into the PrimarySurfaceDesc component.
    ok &= PrimarySurface->GetSurfaceDesc( PrimarySurfaceDesc );

    // retrieve the Back Buffer
    ok &= PrimarySurface->GetAttachedSurface( BackBufferCaps, BackBuffer );

    // Load and Set the palette for the primary surface
    if (BITS == 8)
        {
        ok &= PrimarySurfacePalette->LoadFromFile( Dx_Draw1, "all.bmp" );
        ok &= PrimarySurface->SetPalette( PrimarySurfacePalette );
        }

    // create an offscreen surface and load a bitmap into it
    ok &= TempSurface->CreateFromFile( Dx_Draw1, "all.bmp" );

    // set the primary surface color key
    PrimarySurfaceColorKey->ColorSpaceLowValue = ColorToRGB(clBlack);
    PrimarySurfaceColorKey->ColorSpaceHighValue = RGB(0,0,0);
    ok &= TempSurface->SetColorKey( DDCKEY_SRCBLT, PrimarySurfaceColorKey );
    }
else
    {
    ok = false;
    Screen->Cursor = crDefault;

    AnsiString msg = "This demo cannot be executed on your computer because there is no 640x480x"+IntToStr(BITS)+" display mode available.";
    ShowMessage( msg );
    StatusListBox->Items->Add( "" );
    StatusListBox->Items->Add( "<<Execution Error:>>" );
    StatusListBox->Items->Add( msg );
    }

// Setup Application Events
if (ok)
    {
    Application->OnIdle = OnIdle;
    Application->OnMinimize = OnMinimize;
    Application->OnRestore = OnRestore;
    Active = true;      // used to determine whether animating or not
    }
else
    {
    Top=0;
    Left=0;
    Label1->Visible = true;
    Label2->Visible = true;
    StatusListBox->Top = 0;
    StatusListBox->Left = 0;
    StatusListBox->Width = Width-10;
    StatusListBox->Height = Height-10;
    StatusListBox->Items->Text = "Hello.\n\nA problem has occurred inside this application that means it cannot run as expected.\nAll status messages generated by this application have been listed below for perusal (click, then scroll).\n\nTo EXIT, press ALT-F4.\n--------------------\n" + StatusListBox->Items->Text;
    StatusListBox->Visible = true;
    Active = false;     // used to determine whether animating or not
    }
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::FormDestroy(TObject *Sender)
{
// reset application-specific events
Application->OnMinimize = NULL;
Application->OnRestore = NULL;
Application->OnIdle = NULL;

// destroy various objects (including the backbuffer - it's got internals)
if (BackBuffer->Created) BackBuffer->Destroy();
if (PrimarySurface->Created) PrimarySurface->Destroy();
if (Dx_Draw1->Created)
    {
    Dx_Draw1->RestoreDisplayMode();
    Dx_Draw1->Destroy();
    }
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::OnMinimize( TObject* Sender )
{
// turn off animation when minimized
Active = false;
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::OnRestore( TObject* Sender )
{
// this is one way to successfully restore your application.
bool ok = true;
if (Dx_Draw1->TestCooperativeLevel() != (DDSCL_FULLSCREEN | DDSCL_EXCLUSIVE))
    {
    ok &= Dx_Draw1->SetCooperativeLevel( Application->Handle, DDSCL_FULLSCREEN | DDSCL_EXCLUSIVE );
    ok &= Dx_Draw1->SetDisplayMode( 640, 480, BITS, 0, 0 );
    ok &= Dx_Draw1->SetCooperativeLevel( Application->Handle, DDSCL_FULLSCREEN | DDSCL_EXCLUSIVE );
    }
// only animate if everything worked
Active = ok;
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::OnIdle(TObject *Sender, bool &Done)
{
// if animating, update the relevant frame.
if (Active) UpdateFrame();
// come back to this method as soon as possible.
Done=!Active;
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::UpdateFrame()
{
// this animation function has been implemented with most of the original
// code, I have just changed various names, etc. to make it more understandable

static dword        lastTickCount[3] =  {0, 0, 0};
static int          currentFrame[3] =   {0, 0, 0};
dword               delay[3] =          {50, 78, 13};
int                 xpos[3] =           {288, 190, 416};
int                 ypos[3] =           {128, 300, 256};
int                 i;
dword               thisTickCount;
TRect               rect;

// Decide which frame will be blitted next
thisTickCount = GetTickCount();
for ( i=0; i<3; i++)
    {
    if ((thisTickCount - lastTickCount[i]) > delay[i])
        {
        // Move to next frame;
        lastTickCount[i] = thisTickCount;
        currentFrame[i]++;
        if (currentFrame[i] > 59)
            currentFrame[i] = 0;
        }
    }

// Blit the background for the next frame
rect.Top = 0;
rect.Left = 0;
rect.Right = 640;
rect.Bottom = 480;

while (true)
    {
    BackBuffer->BltFast(0, 0, TempSurface, &rect, DDBLTFAST_NOCOLORKEY );
    if (BackBuffer->ErrorValue == DD_OK) break;
    // the following line of code is no longer required: see the OnSurfaceLost() event below.
    //    if (BackBuffer->ErrorValue == DDERR_SURFACELOST) { if (!RestoreAll()) return; }
    else if (BackBuffer->ErrorValue == DDERR_WASSTILLDRAWING) return;
    }

// Blit the current iteration for the various donuts
for (i=0; i<3; i++)
    {
    rect.Left = currentFrame[i] % 10 * 64;
    rect.Top = currentFrame[i] / 10 * 64 + 480;
    rect.Right = currentFrame[i] % 10 * 64 + 64;
    rect.Bottom = currentFrame[i] / 10 * 64 + 64 + 480;

    while (true)
        {
        BackBuffer->BltFast( xpos[i], ypos[i], TempSurface, &rect, DDBLTFAST_SRCCOLORKEY );
        if (BackBuffer->ErrorValue == DD_OK) break;
        // the following line of code is no longer required: see the OnSurfaceLost() event below.
        //        if (BackBuffer->ErrorValue == DDERR_SURFACELOST) { if (!RestoreAll()) return; }
        else if (BackBuffer->ErrorValue == DDERR_WASSTILLDRAWING) return;
       }
    }

// using standard Win32 display functions, display some text
HDC hdc;
if (BackBuffer->GetDC( &hdc ))
    {
    AnsiString temp_str;

    SetBkColor(hdc, RGB(0, 0, 255));
    SetTextColor(hdc, RGB(255, 255, 0));

    temp_str = "Ticks: "+IntToStr(FrameCount);
    TextOut(hdc, 0, 0, temp_str.c_str(), temp_str.Length() );

    temp_str = "Frames: "+IntToStr(FrameRate);
    TextOut(hdc, 100, 0, temp_str.c_str(), temp_str.Length() );

    BackBuffer->ReleaseDC( hdc );
    }

// using BCB-style paint functions, display some more text
if (BackBuffer->GetDC( &hdc ))
    {
    TCanvas* canvas = new TCanvas();
    if (canvas!=NULL)
        {
        canvas->Handle = hdc;
        canvas->Font->Color = clYellow;
        canvas->Brush->Color = clBlue;
        canvas->Font->Name = "System";
        canvas->TextOut( 500, 0, "Mouse: "+IntToStr(LastMouseX)+","+IntToStr(LastMouseY) );
        canvas->TextOut( 0, canvas->ClipRect.Bottom-60, "Press F3 to save the screen to a file" );
        canvas->TextOut( 0, canvas->ClipRect.Bottom-40, "Press F2 to hide/show Mouse Cursor" );
        canvas->TextOut( 0, canvas->ClipRect.Bottom-20, "Press F1 to minimize" );
        delete canvas;
        }
    BackBuffer->ReleaseDC( hdc );
    }

// display all the changes on the primary surface
while (true)
    {
    PrimarySurface->Flip( NULL, 0 );
    if (PrimarySurface->ErrorValue == DD_OK) break;
    // the following line of code is no longer required: see the OnSurfaceLost() event below.
    //    if (PrimarySurface->ErrorValue == DDERR_SURFACELOST) { if (!RestoreAll()) return; }
    else if (PrimarySurface->ErrorValue == DDERR_WASSTILLDRAWING) return;
    }

// update frame information
FrameCount++;
FrameTime = timeGetTime();
if (FrameTime - LastFrameTime > 1000)
    {
    FrameRate = (FrameCount - LastFrameCount) * 1000 / (FrameTime - LastFrameTime);
    LastFrameTime = FrameTime;
    LastFrameCount = FrameCount;
    }
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::FormKeyDown(TObject *Sender, WORD &Key,
      TShiftState Shift)
{
if (Key==VK_F1) Application->Minimize();
if ((Key==VK_ESCAPE) || (Key==VK_F12)) Close();
if (Key==VK_F2)
    if (Screen->Cursor==crNone) Screen->Cursor = crDefault; else Screen->Cursor = crNone;
if (Key==VK_F3)
    {
    Active = false;

    if (FileExists( "ExampleQ_Snapshot.bmp" ))
        DeleteFile( "ExampleQ_Snapshot.bmp" );

    if (PrimarySurface->SaveToFile( Dx_Draw1, "ExampleQ_Snapshot.bmp" ))
        ShowMessage("Screen Snapshot saved to file :- \n\nDirectory: '"+GetCurrentDir()+"'\nFile: 'ExampleQ_Snapshot.bmp'");
    else
        ShowMessage("Error: Could not save Screen Snapshot to file :- \n\nDirectory: '"+GetCurrentDir()+"'\nFile: 'ExampleQ_Snapshot.bmp'");

    Active = true;
    }
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::PrimarySurfaceSurfaceLost(TObject *Sender)
{
// I've made all 3 the various OnSurfaceLost() events end up here.
if (PrimarySurface->IsLost())
    {
    PrimarySurface->Restore();
    if (BITS==8)
        {
        PrimarySurfacePalette->LoadFromFile( Dx_Draw1, "all.bmp" );
        PrimarySurface->SetPalette( PrimarySurfacePalette );
        }
    }
if (TempSurface->IsLost())
    {
    TempSurface->CreateFromFile( Dx_Draw1, "all.bmp" );
    TempSurface->SetColorKey( DDCKEY_SRCBLT, PrimarySurfaceColorKey );
    }
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::FormMouseMove(TObject *Sender, TShiftState Shift,
      int X, int Y)
{
LastMouseX = X;
LastMouseY = Y;
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::Dx_Draw1Error(TObject *Sender,
      AnsiString TheFunction, AnsiString TheError,
      AnsiString TheErrorMessage)
{
if (StatusListBox->Items->Count>1000) StatusListBox->Items->Clear();

if (TheError!="DDERR_WASSTILLDRAWING")
    {
    StatusListBox->Items->Add( "Error> "+TheFunction );
    StatusListBox->Items->Add( "       > "+TheError );
    StatusListBox->Items->Add( "       > "+TheErrorMessage );
    }
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::FormResize(TObject *Sender)
{
StatusListBox->Width = Width - StatusListBox->Left*2;
StatusListBox->Height = Height - StatusListBox->Top*2;
}
//---------------------------------------------------------------------------

void __fastcall TMainForm::Dx_Draw1EnumDisplayModes(TObject *Sender,
      TDDSurfaceDesc *SurfaceDesc, void *Context, bool &Finished)
{
if (StatusListBox->Items->Count>1000) StatusListBox->Items->Clear();

StatusListBox->Items->Add( "" );
StatusListBox->Items->Add( "Display Mode #" + IntToStr(NumDisplayModes++) );
StatusListBox->Items->Add( "  Width=="+IntToStr(SurfaceDesc->Width) );
StatusListBox->Items->Add( "  Height=="+IntToStr(SurfaceDesc->Height) );
StatusListBox->Items->Add( "  BitCount=="+IntToStr(SurfaceDesc->PixelFormat->RGBBitCount[0]) );

if ((SurfaceDesc->Width == 640) &&
    (SurfaceDesc->Height == 480) &&
    (SurfaceDesc->PixelFormat->RGBBitCount[0] == BITS))
    {
    ModeExists = true;
    Finished = true;
    }
}
//---------------------------------------------------------------------------

